# -*- coding: utf-8 -*-

from plot import *

def plotError ( name, v, N, case='mean', frame=False, color=None, refr=None, K=1, infolines=False, save=None ):
  
  reft = len(refr.times)-1
  refd = refr.var(v, reft, case)
  
  errors = []
  ncells = []
  rtimes = []
  
  for i in range(len(N)):
    
    refdd = refd
    
    zoom = refr.i.NX/N[i][0]
    if refr.i.NY == 1:
      while zoom > 1:
        refdd = (refdd[0:refr.i.NX-1:2,:,:]+refdd[1:refr.i.NX:2,:,:])/2
        zoom /= 2
    elif refr.i.NZ == 1:
      while zoom > 1:
        refdd = (refdd[0:refr.i.NX-1:2,0:refr.i.NY-1:2,:]+refdd[1:refr.i.NX:2,0:refr.i.NY-1:2,:]+refdd[1:refr.i.NX:2,1:refr.i.NY:2,:]+refdd[0:refr.i.NX-1:2,1:refr.i.NY:2,:])/4
        zoom /= 2
    else:
      while zoom > 1:
        refdd = (  refdd[0:refr.i.NX-1:2, 0:refr.i.NY-1:2, 0:refr.i.NZ-1:2] \
                 + refdd[0:refr.i.NX-1:2, 0:refr.i.NY-1:2, 1:refr.i.NZ:2  ] \
                 + refdd[0:refr.i.NX-1:2, 1:refr.i.NY:2,   0:refr.i.NZ-1:2] \
                 + refdd[0:refr.i.NX-1:2, 1:refr.i.NY:2,   1:refr.i.NZ:2  ] \
                 + refdd[1:refr.i.NX:2,   0:refr.i.NY-1:2, 0:refr.i.NZ-1:2] \
                 + refdd[1:refr.i.NX:2,   0:refr.i.NY-1:2, 1:refr.i.NZ:2  ] \
                 + refdd[1:refr.i.NX:2,   1:refr.i.NY:2,   0:refr.i.NZ-1:2] \
                 + refdd[1:refr.i.NX:2,   1:refr.i.NY:2,   1:refr.i.NZ:2  ] ) / 8
        zoom /= 2
    
    name_base = name + '_%dx%dx%d' % (N[i][0], N[i][1], N[i][2])
    
    error = 0
    rtime = 0
    for k in range(1, K+1):
      r = Run('%s-%d' % (name_base, k))
      
      t = len(r.times)-1
      d = r.var(v, t, case)
      
      error += (100*L1(d, refdd))**2
      
      if 'wall_runtime_total' in r.runtime:
        rtime += r.runtime['wall_runtime_total']# - container['mpi_wall_runtime_total']
      else:
        rtime += r.runtime['clock_runtime']
    
    errors.append ( numpy.sqrt(error/K) )
    ncells.append ( N[i][0] )
    rtimes.append ( rtime/K )
  
  # plot error vs ncells
  
  pylab.subplot(121)
  pylab.plot(numpy.log10(ncells), numpy.log10(errors), color+'-x', label=name)
  pylab.title('Relative L1 error of ' + case + ' of ' + v.name + ' (K=%d)' % K)
  pylab.ylabel('log10(error)')
  pylab.xlabel('log10(cells) in x-direction')
  pylab.legend(loc='lower left')
  if not frame: pylab.grid()
  
  slopes = [[1,2], [1,1]]
  '''
  slopes = []
  if case == 'mean':     slopes = [[1,2], [1,1]]
  if case == 'variance': slopes = [[1,4], [1,2]]
  '''
  if not frame:
    for slope in slopes:
      plotSlope(slope, pylab.gca())
  
  # plot error vs runtime
  
  pylab.subplot(122)
  pylab.plot(numpy.log10(rtimes), numpy.log10(errors), color+'-x', label=name)
  pylab.title('Relative L1 error of ' + case + ' of ' + v.name + ' (K=%d)' % K)
  pylab.ylabel('log10(error)')
  pylab.xlabel('log10(seconds)')
  pylab.legend(loc='lower left')
  if not frame: pylab.grid()
  
  dim = int(refr.i.NX>1) + int(refr.i.NY>1) + int(refr.i.NZ>1)
  
  slopes = [[1,2+2*dim], [1,1+dim]]
  
  '''
  if case == 'mean':     slopes = [[1,2+2*dim], [1,1+dim]]
  if case == 'variance': slopes = [[1,4+4*dim], [1,2+2*dim]]
  '''
  if not frame:
    for slope in slopes:
      plotSlope(slope, pylab.gca())
  
  pylab.subplots_adjust(right=0.98)
  pylab.subplots_adjust(left=0.06)
  pylab.subplots_adjust(top=0.94)
  
  if infolines: refr.showInfo('Reference ')
  
  if not frame:
    pylab.show()
  if save:
    pylab.savefig(save)

# plots slopes
def plotSlope(slope, plt):
  
  from pylab import xlim, ylim, loglog
  
  xmin, xmax = xlim()
  ymin, ymax = ylim()
  
  x1 = xmax - (xmax-xmin)/3
  x2 = xmax - (xmax-xmin)/15
  
  y1 = ymax - (ymax-ymin)/20
  y2 = y1 - float(slope[0])/slope[1] * (x2 - x1)
  
  plt.plot([x1, x2], [y1, y2], '--k', label='_nolegend_')
  plt.annotate(' %d/%d' % (slope[0], slope[1]), xy=(x2, y2),  xycoords='data')

def plotEfficiency ( name, N, frame=False, color=None, K=1, save=None ):
  
  ncells = []
  
  clock_runtimes          = []
  wall_runtimes_total     = []
  mpi_wall_runtimes_total = []
  mpi_scalings            = []
  
  for i in range(len(N)):
    
    name_base = name + '_%dx%dx%d' % (N[i][0], N[i][1], N[i][2])
    
    clock_runtime          = 0
    wall_runtime_total     = 0
    mpi_wall_runtime_total = 0
    mpi_scaling            = 0
    
    for k in range(1, K+1):
      container = {}
      execfile('%s_%s-%d.py' % ('runtime', name_base, k), globals(), container)
      clock_runtime          += container['clock_runtime']
      wall_runtime_total     += container['wall_runtime_total']
      mpi_wall_runtime_total += container['mpi_wall_runtime_total']
      mpi_scaling            += container['mpi_scaling']
    
    ncells.append ( N[i][0] * N[i][1] * N[i][2] )
    
    clock_runtimes          .append ( clock_runtime/K )
    wall_runtimes_total     .append ( wall_runtime_total/K )
    mpi_wall_runtimes_total .append ( mpi_wall_runtime_total/K )
    mpi_scalings            .append ( mpi_scaling/K )
  
  # runtimes vs ncells
  
  pylab.subplot(121)
  pylab.plot(numpy.log10(ncells), numpy.log10(wall_runtimes_total), color+'-x', label=name)
  pylab.plot(numpy.log10(ncells), numpy.log10(mpi_wall_runtimes_total), color+'--')
  pylab.title('Runtime: Wall time vs. MPI calls wall time | average of %d runs' % K)
  pylab.ylabel('log10(seconds)')
  pylab.xlabel('log10(cells)')
  pylab.legend(loc='upper left')
  if not frame: pylab.grid()
  
  # MPI efficiency vs ncells
  
  pylab.subplot(122)
  pylab.plot(numpy.log10(ncells), mpi_scalings, color+'-x', label=name)
  pylab.ylim(0,1)
  pylab.title('MPI efficiency: 1 - (MPI wall time) / (wall time) | average of %d runs' % K)
  pylab.ylabel('efficiency')
  pylab.xlabel('log10(cells)')
  pylab.legend(loc='lower right')
  if not frame: pylab.grid()
  
  pylab.subplots_adjust(right=0.98)
  pylab.subplots_adjust(left=0.06)
  pylab.subplots_adjust(top=0.94)
  
  if not frame:
    pylab.show()
  if save:
    pylab.savefig(save)

def plotScaling ( name, N, frame=False, color=None, K=1, save=None ):
  
  # === strong scaling
  
  wall_runtimes_total     = []
  nodes                   = []
  
  for i in range(1, len(N)+1):
    
    name_base = name + '_%d' % i
    
    wall_runtime_total     = 0
    
    for k in range(1, K+1):
      container = {}
      execfile('%s_%s-%d.py' % ('runtime', name_base, k), globals(), container)
      wall_runtime_total     += container['wall_runtime_total']
    
    wall_runtimes_total     .append ( wall_runtime_total/K )
    nodes                   .append ( container['nodes'] )
  
  pylab.subplot(121)
  pylab.plot(numpy.log10(nodes), wall_runtimes_total, color+'-x', label=name)
  (ymin, ymax) = pylab.ylim()
  pylab.ylim(0,ymax)
  pylab.title('Strong scaling | average of %d runs' % K)
  pylab.ylabel('cumulative wall runtime')
  pylab.xlabel('log10(nodes)')
  pylab.legend(loc='upper left')
  if not frame: pylab.grid()
  
  # === weak scaling
  
  wall_runtimes_total     = []
  nodes                   = []
  
  for i in range(len(N)):
    
    name_base = name + '_%dx%dx%d' % (N[i][0], N[i][1], N[i][2])
    
    wall_runtime_total     = 0
    
    for k in range(1, K+1):
      container = {}
      execfile('%s_%s-%d.py' % ('runtime', name_base, k), globals(), container)
      wall_runtime_total     += container['wall_runtime_total']
    
    r = Run( '%s-%d' % (name_base, 1) )
    dim = int(r.i.NX>1) + int(r.i.NY>1) + int(r.i.NZ>1)
    order = int(r.vars['ORDER'])
    
    log_factor = 1
    if dim+1 == order:
      log_factor = ( i + numpy.log2(N[0][0]) ) / numpy.log2(N[0][0])
    
    wall_runtimes_total     .append ( (wall_runtime_total/K) / ((2**(dim+1))**i * log_factor) )
    nodes                   .append ( container['nodes'] )
  
  pylab.subplot(122)
  pylab.plot(numpy.log10(nodes), wall_runtimes_total, color+'-x', label=name)
  (ymin, ymax) = pylab.ylim()
  pylab.ylim(0,ymax)
  pylab.title('"Weak" scaling | average of %d runs' % K)
  pylab.ylabel('cumulative wall time / problem size')
  pylab.xlabel('log10(nodes)')
  pylab.legend(loc='upper left')
  if not frame: pylab.grid()
  
  pylab.subplots_adjust(right=0.98)
  pylab.subplots_adjust(left=0.06)
  pylab.subplots_adjust(top=0.94)
  
  if not frame:
    pylab.show()
  if save:
    pylab.savefig(save)
