# -*- coding: utf-8 -*-

import pylab
import numpy

import os

show = pylab.show

log10_for_variance_treshhold = 100
colors = ['b', 'g', 'r', 'c', 'm', 'y', 'k']

class Variable (object):
    def __init__(self, i, name):
        self.i = i
        self.name = name

RHO = Variable(0, 'rho')
MX = Variable(1, 'mx')
MY = Variable(2, 'my')
MZ = Variable(3, 'mz')
E = Variable(4, 'E')
BX = Variable(5, 'Bx')
BY = Variable(6, 'By')
BZ = Variable(7, 'Bz')
P = Variable(8, 'p')
UX = Variable(9, 'ux')
UY = Variable(10, 'uy')
UZ = Variable(11, 'uz')
A = Variable(12, 'a')
CFX = Variable(13, 'cfx')
CFY = Variable(14, 'cfy')
CFZ = Variable(15, 'cfz')
NVARS = 16

class Config (object):
    def __init__(self, run, conv):
        self.run = run
        self.conv = conv

    def __getattr__(self, n):
        return self.conv(self.run.vars[n])

class Run (object):
    def __init__(self, name=None):
        if name:
            self.outputf = 'output_'+name+'.dat'
            self.varf = 'var_output_'+name+'.dat'
            self.timesf = 'times_'+name+'.py'
            self.configf = 'config_'+name+'.py'
            self.runtimef = 'runtime_'+name+'.py'
            self.movief = 'movie_'+name
            self.name = name
        else:
            self.outputf = 'output.dat'
            self.varf = 'var_output.dat'
            self.timesf = 'times.py'
            self.configf = 'config.py'
            self.runtimef = 'runtime.py'
            self.movief = 'movie'
            self.name = 'run'
        times = {}
        execfile(self.timesf, globals(), times)
        self.times = times['time']
        config = {}
        execfile(self.configf, globals(), config)
        self.opts = config['opts']
        self.vars = config['used']
        self.runtime = {}
        execfile(self.runtimef, globals(), self.runtime)
        self.i = Config(self, int)
        self.f = Config(self, float)
        self.nx = self.i.NX
        self.ny = self.i.NY
        self.nz = self.i.NZ
        
        # extracting the data to be displayed on the plots
        varss = config['vars']
        if 'RUNTIME_FILE' in varss: del varss ['RUNTIME_FILE']
        if 'LOG_FILE'     in varss: del varss ['LOG_FILE']
        if 'TIME_FILE'    in varss: del varss ['TIME_FILE']
        if 'DATA_FILE'    in varss: del varss ['DATA_FILE']
        
        self.s_opts = str(self.opts ).replace('\'','').replace('{','').replace('}','')
        self.s_vars = str(varss).replace('\'','').replace('{','').replace('}','')
    
    def var(self, v, ts, case="mean"):
        if ts > len(self.times):
            raise Exception, "Timestep out of bounds"
        if case == "mean":
            f = open(self.outputf, 'rb')
        else: # case="variance"
            f = open(self.varf, 'rb')
        n = self.nx*self.ny*self.nz
        f.seek(8*n*(v.i + 16*ts))
        fd = numpy.fromfile(f, dtype=numpy.float64, count=n)
        f.close()
        fd.shape = (self.nx, self.ny, self.nz)
        return fd
    
    def showInfo(self, prefix='', adjust=True):
      
      text  = prefix + 'OPTS: ' + self.s_opts + '\n'
      text += prefix + 'VARS: ' + self.s_vars + '\n'
      
      if 'wall_runtime' in self.runtime:
        runtime = self.runtime['wall_runtime']
      else:
        runtime = self.runtime['clock_runtime']
      
      import datetime
      runtime = str(datetime.timedelta(seconds=round(runtime)))
      
      text += prefix + 'INFO: ' + 'nodes:  ' + str(self.runtime['nodes']) + ', ' + 'runtime: ' + runtime
      
      pylab.figtext(0.01, 0.015, text, fontsize='small')
      
      if adjust:
        pylab.subplots_adjust(bottom=0.17)
    
    def plot(self, v, frame=False, ts=None, extent=None, infolines=True, save=None):
        if ts is None: ts = len(self.times)-1
        if self.ny == 1:
            self.plot_1d(v, ts, extent, frame, infolines, save)
        else:
            self.subplot_2d(v, ts, extent, extent, frame, infolines, save)
    
    def plot0(self, v, frame=False, ts=None, extent=None, infolines=True, save=None):
        if not int(self.vars['SAVE0']):
            raise Exception, "Snapshot at t=0 is not available! Use SAVE0=1 when compiling."
        
        if ts is None: ts = len(self.times)-1
        
        if self.ny == 1:
            if not frame:
              pylab.figure(figsize=(2*8,6))
            pylab.subplot(121)
            self.plot_1d(v, 0, extent, True, False)
            pylab.subplot(122)
            self.plot_1d(v, ts, extent, frame, infolines, save)
        else:
            self.subplot_2d(v, 0, extent,  extent, frame, infolines, save)
            self.subplot_2d(v, ts, extent, extent, frame, infolines, save)
    
    def subplot(self, v, frame=False, ts=None, extent_mean=None, extent_var=None, infolines=True, save=None, movie=False):
        if ts is None: ts = len(self.times)-1
        if self.ny == 1:
            self.subplot_1d(v, ts, extent_mean, extent_var, frame, infolines, save, movie)
        else:
            self.subplot_2d(v, ts, extent_mean, extent_var, frame, infolines, save, movie)

    def plot_1d(self, v, ts, extent=None, frame=False, infolines=True, save=None, x=None, y=None, z=None):
        xr = numpy.arange(self.nx)*(self.f.MAXX/self.nx)
        mean = self.var(v, ts, "mean")
        pylab.plot(xr, mean[:,0,0], label="mean")
        
        if not 'COMPUTE_VARIANCE' in self.vars or int(self.vars['COMPUTE_VARIANCE']) == 1:
          var = self.var(v, ts, "variance")
          std = numpy.sqrt(var);
          upper_std = mean + std;
          lower_std = mean - std;
          pylab.plot(xr, upper_std[:,0,0], "--", label="mean + std")
          pylab.plot(xr, lower_std[:,0,0], "--", label="mean - std")
        
        if extent:
            pylab.ylim(*extent)
        pylab.title(v.name + ' t=%.2f' % self.times[ts])
        pylab.legend()
        
        pylab.subplots_adjust(right=0.96)
        pylab.subplots_adjust(left=0.06)
        pylab.subplots_adjust(top=0.94)
        
        if infolines: self.showInfo()
        
        if not frame:
            pylab.show()
        if save:
            pylab.savefig(save)
    
    def subplot_1d(self, v, ts, extent_mean=None, extent_var=None, frame=False, infolines=True, save=None, movie=False, x=None, y=None, z=None):
        
        if movie:
          pylab.figure(figsize=(2*8,6))
        
        xr = numpy.arange(self.nx)*(self.f.MAXX/self.nx)
        
        d = self.var(v, ts, 'mean')
        pylab.subplot(121)
        pylab.plot(xr, d[:,0,0], label=self.name)
        if extent_mean:
            pylab.ylim(*extent_mean)
        pylab.title('mean' + ' of ' + v.name + ' t=%.2f' % self.times[ts])
        pylab.legend()
        
        if not 'COMPUTE_VARIANCE' in self.vars or int(self.vars['COMPUTE_VARIANCE']) == 1:
          d = self.var(v, ts, 'variance')
          pylab.subplot(122)
          pylab.plot(xr, d[:,0,0], label=self.name)
          if extent_var:
              pylab.ylim(*extent_var)
          if not 'COMPUTE_KURTOSIS' in self.vars or int(self.vars['COMPUTE_KURTOSIS']) == 0:
            pylab.title('variance' + ' of ' + v.name + ' t=%.2f' % self.times[ts])
          else:
            pylab.title('4th centered moment' + ' of ' + v.name + ' t=%.2f' % self.times[ts])
          pylab.legend()
        
        pylab.subplots_adjust(right=0.98)
        pylab.subplots_adjust(left=0.03)
        pylab.subplots_adjust(top=0.94)
        
        if infolines: self.showInfo()
        
        if not frame and not movie:
            pylab.show()
        if save:
            pylab.savefig(save)
    
    def subplot_2d(self, v, ts, extent_mean=None, extent_var=None, frame=False, infolines=True, save=None, movie=False):
        
        if not frame:
          pylab.figure(figsize=(13,6))
        
        # mean
        s = {}
        if extent_mean:
          s['vmin'] = extent_mean[0]
          s['vmax'] = extent_mean[1]
        
        d = self.var(v, ts, 'mean')
        pylab.subplot(121)
        
        pylab.imshow(numpy.transpose(d[:,:,0]), origin='lower', extent=(0,self.f.MAXX,0,self.f.MAXY), **s)
        pylab.title('mean' + ' of ' + v.name + ' t=%.2f' % self.times[ts])
        
        # MUST be before all 'pylab.colorbar()' or otherwise plot is distorted (any idea why?)
        if infolines: self.showInfo()
        pylab.subplots_adjust(right=1.00)
        pylab.subplots_adjust(left=0.00)
        pylab.subplots_adjust(top=0.94)
        pylab.subplots_adjust(wspace=0.05)
        
        pylab.colorbar()
        
        # variance
        if not 'COMPUTE_VARIANCE' in self.vars or int(self.vars['COMPUTE_VARIANCE']) == 1:
          s = {}
          if extent_var:
            s['vmin'] = extent_var[0]
            s['vmax'] = extent_var[1]
          
          d = self.var(v, ts, 'variance')
          pylab.subplot(122)
          
          if extent_var:
            pylab.imshow(numpy.transpose(d[:,:,0]), origin='lower', extent=(0,self.f.MAXX,0,self.f.MAXY), **s)
            pylab.title('variance' + ' of ' + v.name + ' t=%.2f' % self.times[ts])
          
          elif numpy.max(d[:,:,0]) >= log10_for_variance_treshhold:
            pylab.imshow(numpy.log10(numpy.transpose(d[:,:,0])), origin='lower', extent=(0,self.f.MAXX,0,self.f.MAXY), **s)
            pylab.title('log10 of ' + 'variance' + ' of ' + v.name + ' t=%.2f' % self.times[ts])
          
          else:
            pylab.imshow(numpy.transpose(d[:,:,0]), origin='lower', extent=(0,self.f.MAXX,0,self.f.MAXY), **s)
            pylab.title('variance' + ' of ' + v.name + ' t=%.2f' % self.times[ts])
          
          pylab.colorbar()
        
        if not frame and not movie:
          pylab.show()
        if save:
          pylab.savefig(save)
    
    def getExtents (self, v, ts, margin=0.0):
      mean = self.var(v, ts, "mean")
      
      low  = numpy.min(mean)
      high = numpy.max(mean)
      diff = high - low
      
      low  -= diff * margin
      high += diff * margin
      
      ext_mean = (low, high)
      
      if not 'COMPUTE_VARIANCE' in self.vars or int(self.vars['COMPUTE_VARIANCE']) == 1:
        var  = self.var(v, ts, "variance")
        low  = numpy.min(var)
        high = numpy.max(var)
        diff = high - low
        
        low  -= diff * margin
        high += diff * margin
      
      ext_var = (low, high)
      
      return [ext_mean, ext_var]
    
    def movie(self, v, slowdown=5, infolines=True, last_extent=True):
        
        margin = 0.1
        
        # for 2d plots, colorbars also indicate both extrema, hence should not be modified
        if self.i.NY != 1:
          margin = 0.0
        
        if last_extent:
          [ext_mean, ext_var] = self.getExtents(v, len(self.times)-1, margin)
        else:
          [ext_mean, ext_var] = self.getExtents(v, 0, margin)
          
          for ts in range(1,len(self.times)):
              
              [ext_mean_tmp, ext_var_tmp] = self.getExtents(v, ts, margin)
              
              ext_mean = (min(ext_mean_tmp[0], ext_mean[0]), max(ext_mean_tmp[1], ext_mean[1]))
              ext_var  = (min(ext_var_tmp [0], ext_var [0]), max(ext_var_tmp [1], ext_var [1]))
        
        for ts in range(len(self.times)):
            self.subplot(v, False, ts, ext_mean, ext_var, infolines, None, True)
            pylab.savefig('movie_%04d.png' % ts)
            pylab.close()
        
        fps = round ( int(self.vars['NSAVES']) / float(self.vars['MAXT']) )
        
        os.system('mencoder "mf://*.png" -mf fps=%d -o %s_%s.mp4 -ovc lavc -lavcopts vcodec=mpeg4:vbitrate=800' % (fps/slowdown, self.movief, v.name))
        #os.system('mencoder "mf://*.png" -mf fps=%d -o %s.avi -ovc lavc -lavcopts vcodec=msmpeg4v2:vbitrate=800' % (fps/slowdown, self.movief))
        
        os.system('rm movie_*.png')
    
    def bfield(self, ts=None):
        if ts is None: ts = len(self.times)-1
        vx,vy = (BX, BY)
        fx = self.var(vx, ts)[:,:,0]
        fy = self.var(vy, ts)[:,:,0]
        fieldlines(fx, fy, self.f.MAXX/self.f.NX, self.f.MAXY/self.f.NY)
    
    def save(self, v, ts=None):
        if ts is None: ts = len(self.times)-1
        if self.ny == 1:
            self.save_1d(v, ts)
        else:
            self.save_2d(v, ts)
    
    def save_1d(self, v, ts):
        x = numpy.arange(self.nx)*(self.f.MAXX/self.nx)
        d = self.var(v, ts)
        pylab.plot(x, d[:,0,0], label=self.name)
        pylab.title(v.name+' t='+str(self.times[ts]))
        pylab.legend()
        pylab.savefig(self.name+'.png')
    
    def save_2d(self, v, ts):
        d = self.var(v, ts)
        pylab.figure(figsize=(8,6))
        pylab.imshow(numpy.transpose(d[:,:,0]), origin='lower',
                     extent=(0,self.f.MAXX,0,self.f.MAXY))
        pylab.colorbar()
        pylab.title(v.name+' t='+str(self.times[ts]))
        pylab.savefig(self.name+'.png')
    
    def DivB2d(self, per, ts=None,):
        fmt="%10.1e"
        if ts is None: ts = len(self.times)-1
        f = open(self.outputf, 'rb')
        n = self.nx*self.ny
        f.seek(8*n*(BX.i + 16*ts))
        bx = numpy.fromfile(f, dtype=numpy.float64, count=n)
        f.close()
        bx.shape = (self.nx, self.ny)
        f = open(self.outputf, 'rb')
        n = self.nx*self.ny
        f.seek(8*n*(BY.i + 16*ts))
        by = numpy.fromfile(f, dtype=numpy.float64, count=n)
        f.close()
        by.shape = (self.nx, self.ny)
        if per == 1:
            bxe = numpy.vstack((bx[-1,:],bx,bx[0,:]))
        else:
            bxe = numpy.vstack((bx[0,:],bx,bx[-1,:]))
        by = numpy.transpose(by)
        if per==1:
            bye = numpy.transpose(numpy.vstack((by[-1,:],by,by[0,:])))
        else:
            bye = numpy.transpose(numpy.vstack((by[0,:],by,by[-1,:])))
        div = (bxe[2:self.nx+2,:] - bxe[0:self.nx,:])/(2*self.f.MAXX/self.nx) + (bye[:,2:self.ny+2] - bye[:,0:self.ny])/(2*self.f.MAXY/self.ny)
        divL1 = abs(div).sum()/(self.f.MAXX*self.f.MAXY*reduce((lambda x,y:x*y), div.shape))
        print (r"divL1= %s"% (fmt)) % divL1
    
    def plotLinf(self, v, refd=None, ts=None):
        if ts is None: ts = len(self.times)-1
        d = self.var(v, ts)
        if refd is None:
            max = numpy.max(abs(d))
        else:
            max = numpy.max(abs(d - refd))/numpy.max(abs(refd))
        print (r"Linf= %s"% max)
    
    def plotmin(self, v, ts=None):
        fmt="%10.2e"
        if ts is None: ts = len(self.times)-1
        d = self.var(v, ts)
        min = numpy.min(abs(d))
        print (r"min= %s \\"% (fmt)) % min

def L1(data, refd=None):
    if refd is None:
        n = self.f.MAXX*self.f.MAXY*reduce((lambda x,y:x*y), data.shape)
        return abs(data).sum()/n
    else:
        return abs(data - refd).sum()/abs(refd).sum()

def Linf(data, refd=None):
    if refd is None:
        return numpy.max(abs(data))
    else:
        return numpy.max(abs(data - refd))/numpy.max(abs(refd))

def tableref(r, runref, v):
    fmt="%10.1e"
    scale=1
    t1 = len(r.times)-1
    t2 = len(runref.times)-1
    d = r.var(v, t1)
    refd = runref.var(v, t2)
    zoom = runref.i.NX/r.i.NX
    if runref.i.NY == 1:
        while zoom>1:
            refd = (refd[0:runref.i.NX-1:2,:,:]+refd[1:runref.i.NX:2,:,:])/2
            zoom=zoom/2
    else:
        while zoom>1:
            refd = (refd[0:runref.i.NX-1:2,0:runref.i.NY-1:2,:]+refd[1:runref.i.NX:2,0:runref.i.NY-1:2,:]+refd[1:runref.i.NX:2,1:runref.i.NY:2,:]+refd[0:runref.i.NX-1:2,1:runref.i.NY:2,:])/4
            zoom=zoom/2
    td = (L1(d, refd)*scale, Linf(d, refd)*scale)
    print (r"%s & %s & %s \\"% (r.name,fmt,fmt)) % td

def rebin(d, f):
    nx,ny,nz = d.shape
    if ny == 1:
        assert nz == 1
        return sum(d[i:nx-f+1+i:f,:,:] for i in range(f))/f
    elif nz == 1:
        return sum(d[i:nx-f+1+i:f,i:ny-f+1+i:f,:] for i in range(f))/(f*f)
    else:
        return sum(d[i:nx-f+1+i:f,i:ny-f+1+i:f,i:nz-f+1+i:f] for i in range(f))/(f*f*f)

def table(runs, v, t=None, fmt="%12.1e", refd=None, reft=None, refr=None, scale=1, dt=0.01):
    if t is None:
        t = max(r.times[-1] for r in runs)
    print r'       Run &        $L_1$ &   $L_\infty$ \\'
    for r in runs:
        ts = None
        for n,tm in enumerate(r.times):
            if abs(tm - t) < dt:
                ts = n
                break
        if ts is None:
            print r.name + r' & crashed \\'
            continue
        d = r.var(v, ts)
        if reft is not None:
            refts = None
            for n,tm in enumerate(r.times):
                if tm == reft:
                    refts = n
                    break
            refd = r.var(v, refts)
        elif refr is not None:
            refd = refr.var(v, ts)
            if refr.i.NX != r.i.NX:
                assert (refr.i.NX//r.i.NX) * r.i.NX == refr.i.NX
                refd = rebin(refd, refr.i.NX // r.i.NX)
        td = (L1(d, refd)*scale, Linf(d, refd)*scale)
        print (r"%10s & %s & %s \\"% (r.name,fmt,fmt)) % td

def interpolate(v, x, y):
    x0 = int(numpy.floor(x))
    y0 = int(numpy.floor(y))
    x1 = int(numpy.ceil(x))
    y1 = int(numpy.ceil(y))
    if x1 >= v.shape[0]: x1 = v.shape[0]-1
    if y1 >= v.shape[1]: y1 = v.shape[1]-1
    if x0 == x1:
        if y0 == y1:
            return v[x0,y0]
        return v[x0,y0]*(y-y0) + v[x0,y1]*(y1-y)
    elif y0 == y1:
        return v[x0,y0]*(x-x0) + v[x1,y0]*(x1-x)
    return (v[x0,y0]*(x-x0) + v[x1,y0]*(x1-x))*(y-y0) + \
        (v[x0,y1]*(x-x0) + v[x1,y1]*(x1-x))*(y1-y)

def fieldlines(fx, fy, xsc=1.0, ysc=1.0, nlines=20, colour='k'):
    assert fx.shape == fy.shape
    dfy0 = sum(x for x in fy[:,0] if x > 0) / (nlines+1)
    fy0 = 0
    x0 = 0
    for li in range(nlines):
        while fy0 < dfy0:
            if fy[x0,0] > 0: fy0 += fy[x0,0]
            x0 += 1
        fy0 -= dfy0
        x = x0 - fy0/fy[x0-1,0]
        y = 0
        lx = [xsc*x]
        ly = [ysc*y]
        while x >= 0 and y >= 0 and x <= fx.shape[0] and y <= fx.shape[1]:
            dx = interpolate(fx, x, y)
            dy = interpolate(fy, x, y)
            h = numpy.sqrt(dx*dx+dy*dy)
            x += dx/h
            y += dy/h
            lx.append(xsc*x)
            ly.append(ysc*y)
        pylab.plot(lx, ly, colour)

print ""

try:
  r = Run();
  print " :: Default run loaded as \"r\"."
except:
  print " :: Default run not found!"

try:
  ref = Run('Reference');
  print " :: Reference run loaded as \"ref\"."
except:
  print " :: Reference run not found!"

try:
  mc = Run('MC');
  print " :: MC run loaded as \"mc\"."
except:
  print " :: MC run not found!"

try:
  mlmc = Run('MLMC');
  print " :: MLMC run loaded as \"mlmc\"."
except:
  print " :: MLMC run not found!"

try:
  mlmcd = Run('MLMCd');
  print " :: MLMCd run loaded as \"mlmcd\"."
except:
  print " :: MLMCd run not found!"

try:
  mc2 = Run('MC2');
  print " :: MC2 run loaded as \"mc2\"."
except:
  print " :: MC2 run not found!"

try:
  mlmc2 = Run('MLMC2');
  print " :: MLMC2 run loaded as \"mlmc2\"."
except:
  print " :: MLMC2 run not found!"

try:
  mlmc2d = Run('MLMC2d');
  print " :: MLMC2d run loaded as \"mlmc2d\"."
except:
  print " :: MLMC2d run not found!"

print ""

def tables (var = RHO):
  table([mc, mlmc, mlmcd, mc2, mlmc2, mlmc2d], var, None, "%12.1e", None, None, ref)

def movies (slowdown=5):
  r.movie (RHO, slowdown)
  r.movie (P,   slowdown)
  r.movie (E,   slowdown)
  r.movie (UX,  slowdown)
  r.movie (MX,  slowdown)
  r.movie (BX,  slowdown)