
#ifndef CHAOS_H
#define CHAOS_H

#include <cstdlib>
#include <iostream>

#include "config.h"
#include "WELL512a.h"
#include "primes.h"

struct Chaos {
  // double arrays only!
  double uniform [N_UNIFORM];
  double normal  [N_NORMAL];
};

static const int CHAOS_SIZE = sizeof(Chaos) / sizeof(double);

extern Chaos w;
extern Chaos *iid;
extern unsigned long int iid_size;

static void chaos_resize (unsigned long int M) {
  iid_size = M;
  iid = (Chaos*) realloc (iid, iid_size*sizeof(Chaos));
  if (iid == NULL) {
    std::cout << " :: ERROR: out of memory." << std::endl;
    throw 1;
  }
}

static Chaos generate_chaos () {
  Chaos chaos;
  
  for (int i=0; i<N_UNIFORM; i++)
    chaos.uniform[i] = WELLRNG512a();
  
  for (int i=0; i<N_NORMAL; i++) {
    chaos.normal[i] = WELLRNG512a();
  }
  
  return chaos;
}

static void chaos_init () {
  for (unsigned long int i=0; i<iid_size; i++)
    iid[i] = generate_chaos ();
}

static void chaos_buffer_init (unsigned int seed) {
  unsigned int buffer[getR()];
  
  // init buffer using Linear Congruential Generator
  
  const unsigned long int a = 1103515245;
  const unsigned long int c = 12345;
  
  unsigned int x = seed;
  
  for (int i=0; i<getR(); i++) {
    x = a*x + c;
    buffer[i] = x;
  }
  
  InitWELLRNG512a (buffer);
}

#endif
