
#include <fstream>
#include <sstream>

#include "MHD.h"

class Saver
{
private:  
  std::ofstream *data;
  std::ofstream *times;
  volume <ConservedVars, 1> *CV;
  volume <XtraVars, 1>      *XV;
  bool nonnegative;

public:
  Saver(const std::string quantity="mean");
  ~Saver();
  void step(volume <ConservedVars, 1> &CV, volume <XtraVars, 1> &XV, real t_halt);
  void snapshots();
  void write(real &r) { data->write((char *) &r, sizeof(real)); }
};

Saver::Saver(std::string quantity)
{
  
  std::string prefix;
  
  if (!quantity.compare("mean")) {
    // construct mean Saver
    CV = &mean_U[0];
    XV = &mean_X[0];
    prefix = "";
    nonnegative = false;
  }
  #if COMPUTE_VARIANCE
  else {
    // construct variance Saver
    CV = &var_U[0];
    XV = &var_X[0];
    prefix = "var_";
    if (VARIANCE_FIX)
      nonnegative = true;
    else
      nonnegative = false;
  }
  #endif
  
  std::ostringstream dfn;
  if (MULTIX > 1 || MULTIY > 1 || MULTIZ > 1)
    dfn << prefix << DATA_FILE << "." << multix << "." << multiy << "." << multiz;
  else
    dfn << prefix << DATA_FILE;
  data = new std::ofstream(dfn.str().c_str());
  if (multi_main) {
    times = new std::ofstream(TIME_FILE);
    *times << "time = [";
  }
}

Saver::~Saver()
{
  delete data;
  if (multi_main) {
    *times << "]\n";
    delete times;
  }
}

template<class T, int G> void write_data(Saver *s, volume<T,G> &v, real T::*d)
{
  for (int i = 0; i < NX/MULTIX; i++) {
    for (int j = 0; j < NY/MULTIY; j++) {
      for (int k = 0; k < NZ/MULTIZ; k++) {
	s->write(v[i][j][k].*d);
      }
    }
  }
}

template<class T, int G> void write_data(Saver *s, volume<T,G> &v, vector T::*d)
{
  for (int di = 0; di < NDIR; di++) {
    for (int i = 0; i < NX/MULTIX; i++) {
      for (int j = 0; j < NY/MULTIY; j++) {
	for (int k = 0; k < NZ/MULTIZ; k++) {
	  s->write((v[i][j][k].*d)[di]);
	}
      }
    }
  }
}

void Saver::step(volume <ConservedVars, 1> &CV, volume <XtraVars, 1> &XV, real t_halt)
{
  if (multi_subdomains_root) {
    if (multi_main) {
      *times << t_halt <<", ";
      times->flush();
    }
    write_data(this, CV, &ConservedVars::rho);
    write_data(this, CV, &ConservedVars::m);
    write_data(this, CV, &ConservedVars::E);
    write_data(this, CV, &ConservedVars::B);
    write_data(this, XV, &XtraVars::p);
    write_data(this, XV, &XtraVars::u);
    write_data(this, XV, &XtraVars::a);
    write_data(this, XV, &XtraVars::cf);
    data->flush();
  }
}

void Saver::snapshots ()
{
  for(int s = 0; s<NSAVES+SAVE0; s++) {
    
    // for stabilization of ML-part with exp-log
    CV[s].log();
    XV[s].log();
    
    if (nonnegative) {
      CV[s].nonnegative();
      XV[s].nonnegative();
    }
    
    step(CV[s], XV[s], halt_times[s]);
  }
}
