
#include "case.h"
#include "hll_util.h"
#include "solver.h"

///::title:: HLL5 solver ("second" 5 wave)
///::depends:: hll_util

void solver_calc_flux(ConservedVars &F, vector &us, vector &Bs,
		      const AllVariables &lV, const AllVariables &rV,
		      int i, int j, int k, int d)
{
  real sl, sr, cs;
  ConservedVars lF, rF, Us, Us5,Uss5;
  vector mls5,mrs5,bls5,brs5;
  vector c5,d5;

  real tol = 0.0001;
  
  //  hll_speeds(lV, rV, d, sl, sr, cs);
  hll_speeds_ext(lV, rV, d, 1e-5, sl, sr, cs);
  calc_point_flux(lF, lV, d);
  calc_point_flux(rF, rV, d);

  real ptl = lV.XU.p + 0.5*lV.U.B.sqr();
  real ptr = rV.XU.p + 0.5*rV.U.B.sqr();

  real udl = lV.XU.u[d]-sl;
  real udr = rV.XU.u[d]-sr;

  real aa = udr*rV.U.rho - udl*lV.U.rho;

  real sm = (rV.U.m[d]*udr - lV.U.m[d]*udl + ptr - ptl -
	     (sqr(rV.U.B[d]) - sqr(lV.U.B[d]))/2)/aa;

  real rhols5 = lV.U.rho*udl/(sm-sl);
  real rhors5 = rV.U.rho*udr/(sm-sr);

  real sls5 = sm-fabs(lV.U.B[d])/ssqrt(rhols5);
  real srs5 = sm+fabs(rV.U.B[d])/ssqrt(rhors5);

  real denl5 = lV.U.rho*udl*(sm-sl)-sqr(lV.U.B[d]);
  real denr5 = rV.U.rho*udr*(sm-sr)-sqr(rV.U.B[d]);

  if (fabs(denl5)<=tol){
    mls5 = rhols5*lV.XU.u;
    bls5 = lV.U.B;
  } else {
    mls5=rhols5*(lV.XU.u + lV.U.B[d]*lV.U.B*(lV.XU.u[d]-sm)/denl5);
    bls5=lV.U.B*(lV.U.rho*sqr(udl)-sqr(lV.U.B[d]))/denl5;
  }
  mls5[d] = rhols5*sm;
  bls5[d] = lV.U.B[d];

  if (fabs(denr5)<=tol){
    mrs5 = rhors5*rV.XU.u;
    brs5 = rV.U.B;
  } else {
    mrs5=rhors5*(rV.XU.u + rV.U.B[d]*rV.U.B*(rV.XU.u[d]-sm)/denr5);
    brs5=rV.U.B*(rV.U.rho*sqr(udr)-sqr(rV.U.B[d]))/denr5;
  }
  mrs5[d] = rhors5*sm;
  brs5[d] = rV.U.B[d];

  real alpha = (srs5-sm)*rhors5 + (sm-sls5)*rhols5;
  real beta = rV.U.B[d] - lV.U.B[d];
  real gamma = srs5-sls5;
  //  real agmbb = alpha*gamma - beta*beta;
  real agmbb = fabs(lV.U.B[d])*fabs(rV.U.B[d])*(rhors5+rhols5)/ssqrt(rhors5*rhols5) + 2*rV.U.B[d]*lV.U.B[d];
  int threewaveflag=0;
  //  if (fabs(lV.U.B[d])/ssqrt(rhols5)<=tol||fabs(rV.U.B[d])/ssqrt(rhols5)<=tol) threewaveflag=1;
  if (fabs(lV.U.B[d])<=tol||fabs(rV.U.B[d])<=tol) threewaveflag=1;
  else if (sign(rV.U.B[d])!=sign(lV.U.B[d])&&fabs(rhors5-rhols5)!=0) threewaveflag=1;
  if (threewaveflag) {

    //////// relax to 3 wave solver /////////

  vector Gm = rF.m - lF.m - sr*rV.U.m + sl*lV.U.m;
  vector GB = rF.B - lF.B - sr*rV.U.B + sl*lV.U.B;
  real bb = rV.U.B[d] - lV.U.B[d];
  real cc = sl - sr;
  real acmbb = aa*cc-bb*bb;
  
  us = (cc*Gm + bb*GB)/acmbb;
  us[d] = sm;
  Bs = (bb*Gm + aa*GB)/acmbb;
  Us.B = Bs;


  if (sl > 0) {F = lF; Bs[d] = lV.U.B[d];}
  else if (sr < 0) {F = rF; Bs[d] = rV.U.B[d];}
  else if (sm >= 0) {
    Us.rho = lV.U.rho*udl/(sm-sl);
    Us.m = Us.rho * us;
    Us.B[d] = lV.U.B[d];
    real pts = ptl + lV.U.rho*(lV.XU.u[d]-sm)*udl;
    Us.E = (udl*lV.U.E + ptl*lV.XU.u[d] - pts*sm +
	    lV.U.B[d]*(Us.B*us - lV.U.B*lV.XU.u)) / (sm-sl);
    F = lF + sl*(Us-lV.U);
  } else {
    Us.rho = rV.U.rho*udr/(sm-sr);
    Us.m = Us.rho * us;
    Us.B[d] = rV.U.B[d];
    real pts = ptr + rV.U.rho*(rV.XU.u[d]-sm)*udr;
    Us.E = (udr*rV.U.E + ptr*rV.XU.u[d] - pts*sm +
	    rV.U.B[d]*(Us.B*us - rV.U.B*rV.XU.u)) / (sm-sr);
    F = rF + sr*(Us-rV.U);
  }
  } else {

    /////////// use 5 wave solver ////////////


  c5 = mrs5*(srs5-sm) + mls5*(sm-sls5) + rV.U.B[d]*brs5 - lV.U.B[d]*bls5;
  d5 = (srs5-sm)*brs5 + (sm-sls5)*bls5 + mrs5/rhors5*rV.U.B[d] - mls5/rhols5*lV.U.B[d];

      us = (c5*gamma-d5*beta)/agmbb;
      us[d] = sm;
      Bs = (d5*alpha-c5*beta)/agmbb;

  if (sl > 0) {F = lF; Bs[d] = lV.U.B[d];}
  else if (sr < 0) {F = rF; Bs[d] = rV.U.B[d];}
  else if (sm>0) {
	real pts5 = ptl + lV.U.rho*(lV.XU.u[d]-sm)*udl;
	Us5.rho = rhols5;
	Us5.m = mls5;
	Us5.B = bls5;
	Us5.E = ( udl*lV.U.E + ptl*lV.XU.u[d]-pts5*sm + lV.U.B[d]*(mls5/rhols5*bls5 - lV.XU.u*lV.U.B))/(sm-sl);

// 	  if (Us5.E<=0) {
// 	  printf("neg E(sm>0)");
// 	  }
	F = lF + sl*(Us5-lV.U);
	if (sls5<=0) {
	  Bs[d]=bls5[d];
	  Uss5.rho = rhols5;
	  Uss5.m = rhols5*us;
	  Uss5.B = Bs;
	  Uss5.B[d] = lV.U.B[d];

// 	  vector uls5zerod = mls5/rhols5;
// 	  	  uls5zerod[d]=0;
// 	  vector uszerod = us;
// 	  	  uszerod[d]=0;
// 	  Uss5.E = ( (sls5-sm)*Us5.E + lV.U.B[d]*(uls5zerod*Us5.B-uszerod*Uss5.B) )/(sls5-sm);
	  Uss5.E = ( (sls5-sm)*Us5.E + lV.U.B[d]*(mls5/rhols5*Us5.B-us*Uss5.B) )/(sls5-sm);

// 	  if (Uss5.E<=0) {
// 	  printf("neg E(sm>0&sls5<=0)");
// 	  }
	  F += sls5*(Uss5-Us5);
	}
  } else {
	real pts5 = ptr + rV.U.rho*(rV.XU.u[d]-sm)*udr;
	Us5.rho = rhors5;
	Us5.m = mrs5;
	Us5.B = brs5;
	Us5.E = ( udr*rV.U.E + ptr*rV.XU.u[d]-pts5*sm + rV.U.B[d]*(mrs5/rhors5*brs5 - rV.XU.u*rV.U.B))/(sm-sr);

// 	  if (Us5.E<=0) {
// 	  printf("neg E(sm<=0)");
// 	  }
	F = rF + sr*(Us5-rV.U);
	if (srs5>0) {
	  Bs[d]=brs5[d];
	  Uss5.rho = rhors5;
	  Uss5.m = rhors5*us;
	  Uss5.B = Bs;
	  Uss5.B[d] = rV.U.B[d];
	  
// 	  vector urs5zerod = mrs5/rhors5;
// 	  	  urs5zerod[d]=0;
// 	  vector uszerod = us;
// 	  	  uszerod[d]=0;
// 	  Uss5.E = ( (srs5-sm)*Us5.E + rV.U.B[d]*(urs5zerod*Us5.B-uszerod*Uss5.B) )/(srs5-sm);
	  Uss5.E = ( (srs5-sm)*Us5.E + rV.U.B[d]*(mrs5/rhors5*Us5.B-us*Uss5.B) )/(srs5-sm);

// 	  if (Uss5.E<=0) {
// 	  printf("neg E(sm<=0&srs5>0)");
// 	  }
	    F += srs5*(Uss5-Us5);
	}
  }

}

}
