
#include "case.h"
#include "space.h"
#include "solver.h"
#include "hll_util.h"

// substep call for first and second order

void sub_step(bool firsttime)
{
  PRE_STEP_HOOK
  // calculate fluxes & sources
  for (int i = 0; i < NXl; i++) {
    for (int j = 0; j < NYl; j++) {
      for (int k = 0; k < NZl; k++) {
	Source[i][j][k].zero();
	SOURCE_HOOK
      }
    }
  }

  AllVariables lV, rV;
  vector us, Bs;
  ////
  vector maxeig;
  maxeig.zero();
  real sl, sr, cs;
  ////
  for (int d = 0; d < NDIR; d++) {
    if (!HAVE_DIR(d)) continue;
    for (int i = 0; i < FLUX_NX(d); i++) {
      for (int j = 0; j < FLUX_NY(d); j++) {
	for (int k = 0; k < FLUX_NZ(d); k++) {
	  right(lV, i-DX[d], j-DY[d], k-DZ[d], d);
	  left(rV, i, j, k, d);
	  ///////////
	  if (firsttime) {
	    hll_speeds(lV, rV, d, sl, sr, cs);
	    maxeig[d] = max(maxeig[d],max(fabs(sl),fabs(sr)));
	  }
	  ///////////
	  solver_calc_flux(Flux[d][i][j][k], us, Bs, lV, rV, i, j, k, d);
	  if (POWELL && !NO_POWELL)
	    powell_source(us, Bs, lV, rV, i, j, k, d);
	}
      }
    }
  }

  ///////////
  if (firsttime) {
    dt = multi_dt(CFL/(maxeig[X]/dx[X] + maxeig[Y]/dx[Y] + maxeig[Z]/dx[Z]));
    if (t + dt > t_halt) dt = t_halt - t;
  }
  
  // apply fluxes & sources

  for (int i = 0; i < NXl; i++) {
    for (int j = 0; j < NYl; j++) {
      for (int k = 0; k < NZl; k++) {
	for (int d = 0; d < NDIR; d++) {
	  if (!HAVE_DIR(d)) continue;
	  U[i][j][k] -= (dt/dx[d])*(Flux[d][i+DX[d]][j+DY[d]][k+DZ[d]] -
				    Flux[d][i][j][k]);
	}
	U[i][j][k] += dt * Source[i][j][k];
      }
    }
  }
}
