
#include "case.h"
#include "space.h"
#include "solver.h"

// substep call for 3rd order, including quadrature integration.

void sub_step(bool firsttime)
{
  PRE_STEP_HOOK
  // calculate fluxes & sources
  for (int i = 0; i < NXl; i++) {
    for (int j = 0; j < NYl; j++) {
      for (int k = 0; k < NZl; k++) {
	Source[i][j][k].zero();
	SOURCE_HOOK
      }
    }
  }

  AllVariables lV, rV;
  vector us, Bs;
  for (int d = 0; d < NDIR; d++) {
    if (!HAVE_DIR(d)) continue;
    for (int i = 0; i < FLUX_NX(d); i++) {
      for (int j = 0; j < FLUX_NY(d); j++) {
	for (int k = 0; k < FLUX_NZ(d); k++) {
	  ConservedVars F;
	  Flux[d][i][j][k].zero();
	  for (int p = 0; p < NQUAD; p++) {
	    right(lV, i-DX[d], j-DY[d], k-DZ[d], d, p);
	    left(rV, i, j, k, d, p);
	    solver_calc_flux(F, us, Bs, lV, rV, i, j, k, d);
	    Flux[d][i][j][k] += QUAD_WEIGHT[p]*F;
	    if (POWELL && !NO_POWELL)
	      powell_source(us, Bs, lV, rV, i, j, k, d, 1.0/NQUAD);
	  }
	}
      }
    }
  }

  // apply fluxes & sources

  for (int i = 0; i < NXl; i++) {
    for (int j = 0; j < NYl; j++) {
      for (int k = 0; k < NZl; k++) {
	for (int d = 0; d < NDIR; d++) {
	  if (!HAVE_DIR(d)) continue;
	  U[i][j][k] -= (dt/dx[d])*(Flux[d][i+DX[d]][j+DY[d]][k+DZ[d]] -
				    Flux[d][i][j][k]);
	}
	U[i][j][k] += dt * Source[i][j][k];
      }
    }
  }
}
